using System;
using System.Collections.Generic;
using System.Text;
// ...
using System.Windows.Forms;
// ... extra for serial communication
using System.IO.Ports;
using System.Threading;
using System.IO;

namespace ClassLibrary2
{
    public class ParallaxServoController
        // Class interfacing the Parallax Servo Controller
        //
        // Tested OK, but distributed "as is" 
        //
        // (c) 2006 / Herman Lenferink / www.fscockpit.com
        // Only re-distribute with the original code enclosed
    {
        // The 2 baud rates possible for the Parallax Servo Controller 
        public const byte Baud2400 = 0;
        public const byte Baud38K4 = 1;

        // The Parallax Servo Controller starts at servo position 250 and stops at position 1250
        public const UInt16 ServoStart = 250;
        public const UInt16 ServoSteps = 1250 - ServoStart;

        public string TheCOMPort = "";
        public int ThePCBaudRate ;
                
        // Default timeout after 5 seconds
        public int TheReadTimeout = 5000;
        public int TheWriteTimeout = 5000;

        // Standard serial interface
        private SerialPort Com = new SerialPort();

        // To the Parallax Servo Controller ASCII code (bytes) is send
        private ASCIIEncoding Ascii = new ASCIIEncoding();

        private void ErrorMessage(string Message)
            // Display error message
        {
            MessageBox.Show(Message, "Parallax Servo Controller", MessageBoxButtons.OK, MessageBoxIcon.Error);
        }

        private void WarningMessage(string Message)
            // Display Warning message
        {
            MessageBox.Show(Message, "Parallax Servo Controller", MessageBoxButtons.OK, MessageBoxIcon.Warning);
        }

        private void Pause(double Second)
            // Simple way to pause execution for +/- a certain amount of time
        {
            long StartTick, EndTick, Duration;
            Duration = (long) (Second * 10000000);
            StartTick = DateTime.Now.Ticks;
            do
            {
                EndTick = DateTime.Now.Ticks;
            } while ((EndTick - StartTick) < Duration);
        }

        private void PSCWrite(byte[] Token)
            // Com.Write with all the "catch" needed
        {
            try
            {
                Com.Write(Token, 0, Token.Length);
            }
            catch (TimeoutException)
            {
                ErrorMessage("PSCWrite - TimeoutException: The operation did not complete before the time-out period ended");
            }
            catch (InvalidOperationException)
            {
                ErrorMessage("PSCWrite - InvalidOperationException: The specified port is not open");
            }
        }

        private void PSCRead(ref byte[] Buffer,int Offset,int NumberOfBytes)
            // Com.Read with all the "catch" needed
        {
            try
            {
                Com.Read(Buffer,Offset,NumberOfBytes);
            }
            catch (TimeoutException)
            {
                ErrorMessage("PSCRead - TimeoutException: The operation did not complete before the time-out period ended");
                // or
                // No bytes were read
            }
            catch (InvalidOperationException)
            {
                ErrorMessage("PSCRead - InvalidOperationException: The specified port is not open");
            }
            catch (ArgumentNullException)
            {
                ErrorMessage("PSCRead - ArgumentNullException: The buffer passed is a null reference");
            }
            catch (ArgumentOutOfRangeException)
            {
                ErrorMessage("PSCRead - ArgumentOutOfRangeException: The offset or count parameters are outside a valid region of the buffer being passed. Either offset or count is less than zero.");
            }
            catch (ArgumentException)
            {
                ErrorMessage("PSCRead - ArgumentException: offset plus count is greater than the length of the buffer.");
            }
        }

        private void PSCEmptyReadBuffer()
            // Com.DiscardInBuffer with all the "catch" needed
        {
            try
            {
                Com.DiscardInBuffer();
            }
            catch (IOException)
            {
                ErrorMessage("PSCEmptyReadBuffer - IOException: The port is in an invalid state or an attempt to set the state of the underlying port failed. For example, the parameters passed from this SerialPort object were invalid.");
            }
            catch (InvalidOperationException)
            {
                ErrorMessage("PSCEmptyReadBuffer - IOException: The stream is closed. This can occur because the Open method has not been called or the Close method has been called.");
            }
        }

        private void PSCSetPCBaudRate(byte BaudRate)
            // Set baud rate for PC side
        {
            try
            {
                switch (BaudRate)
                {
                    case Baud2400:
                        Com.BaudRate = 2400;
                        ThePCBaudRate = 2400;
                        break;
                    case Baud38K4:
                        Com.BaudRate = 38400;
                        ThePCBaudRate = 38400;
                        break;
                    default:
                        ErrorMessage("PSCSetPCBaudRate - Wrong PC baud rate.");
                        break;
                }
            }
            catch (ArgumentOutOfRangeException)
            {
                ErrorMessage("PSCSetPCBaudRate - ArgumentOutOfRangeException: Given value out of range");
            }
            catch (IOException)
            {
                ErrorMessage("PSCSetPCBaudRate = IOException: The port is in an invalid state or an attempt to set the state of the underlying port failed. For example, the parameters passed from this SerialPort object were invalid.");
            }
        }

        private void PSCOpenCOM()
            // Open COM port
        {
            if (Com.IsOpen == false)
            {
                try
                {
                    Com.Open();
                }
                catch (ArgumentOutOfRangeException)
                {
                    ErrorMessage("OpenConnection - ArgumentOutOfRangeException: Given value out of range");
                }
                catch (IOException)
                {
                    ErrorMessage("OpenConnection = IOException: The port is in an invalid state or an attempt to set the state of the underlying port failed. For example, the parameters passed from this SerialPort object were invalid.");
                }
                catch (InvalidOperationException)
                {
                    ErrorMessage("OpenConnection - The specified port is open.");
                }
                catch (ArgumentException)
                {
                    ErrorMessage("OpenConnection - The port name does not begin with COM Or the file type of the port is not supported.");
                }
                catch (UnauthorizedAccessException)
                {
                    ErrorMessage("OpenConnection - Access is denied to the port.");
                }
            }
        }

        public void OpenConnection()
        // Open the Parallax Servo Controller connection
        {
            if (Com.IsOpen == false)
            {
                    PSCSetPCBaudRate(Baud2400);
                    ThePCBaudRate = Com.BaudRate;
                    Com.ReadTimeout = TheReadTimeout;
                    Com.WriteTimeout = TheWriteTimeout;
                    Com.PortName = TheCOMPort;

                    PSCOpenCOM();
            }
        }

        public void CloseConnection()
            // Close the COM port
        {
            try
            {
                Com.Close();
            }
            catch (InvalidOperationException)
            {
                ErrorMessage("CloseConnection - The specified port is not open.");
            }
        }

        public string GetVersionNumber()
            // Get version number of the Parallax Servo Controller
        {
            byte[] Token = new byte[8];
            byte[] PSCBuffer = new byte[11];

            // The buffer may not be empty, so this is done explicitly
            PSCEmptyReadBuffer();            
            
            //                      01234567  
            Token = Ascii.GetBytes("!SCVER?\x0D");
            // Syntax: !SCVER? $0D (no spaces)
            // Reply:  3 bytes with version number e.g. "1.3"
            // >>> Actual reply is 8 bytes command and 3 bytes ...

            // ... NOTE ...
            // When 2 Parallax boards are connected 6 characters have to be read

            // Sent command to Parallax Servo Controller
            PSCWrite(Token);

            // Wait for 0.4 seconds to give the PSC the chance to respond
            Pause(0.4);

            // Returns the command (8 positions) and the version number (3 positions)
            PSCRead(ref PSCBuffer, 0, 11);

            // return only the version number
            return Ascii.GetString(PSCBuffer,8,3);
        }

        public void SetBaudRate(byte Rate)
            // Set baud rate for the Parallax Servo Controller
        {
            byte[] Token = new byte[8];
            byte[] PSCBuffer = new byte[11];

            // The buffer may not be empty, so this is done explicitly
            PSCEmptyReadBuffer();

            //                      01234567  
            Token = Ascii.GetBytes("!SCSBR_\x0D");
            // Syntax: !SCSBR x $0D (no spaces) 
            //         Where x = 0 for 2400 or 1 for 38K4
            // Reply:  BR x (no spaces / 3 bytes)
            // >>> Actual reply is 8 bytes command, the 3 reply bytes are not produced ...

            // Set baud rate
            Token[6] = Rate;

            // Sent command to Parallax Servo Controller
            PSCWrite(Token);

            // Read command from buffer, ths is still in old baud rate
            // Ignore the result
            // Actually the result schould 11 bytes with the latter 3 
            // read after "PSCOpenCOM" below
            PSCRead(ref PSCBuffer, 0, 8);

            // In order to swith to ne baud rate the COM connection has to be closed
            CloseConnection();
            
            // Set new baud rate
            PSCSetPCBaudRate(Rate);

            // wait for 0.4 seconds to give the PSC the chance to respond
            Pause(0.4);

            // Open COM connection again
            PSCOpenCOM();
        }

        public void MoveServo(byte Channel, byte Ramp, UInt16 NewPosition)
            // Mover servo to new position
        {
            byte[] NewPositionInBytes = new byte[2];
            byte[] Token = new byte[8];
            byte[] PSCBuffer = new byte[8];

            //                         01234567  
            Token = Ascii.GetBytes("!SCcrlh\x0D");
            // Syntax: !SC C R P.LOWBYTE P.HIGHBYTE $0D
            //         C = servo channel, binary number 1 - 31 (16 - 31 on second board)
            //         R = Ramp value, binary number 0 - 63
            //         P = 16 bit word (2 bytes) that corresponds to desired servo position
            //             min value 250
            //             max value 1250
            //             Note: If servo appears to strain at position 250 or 1250
            //                   250 should be increased or 1250 should be decreased

            // Select servo
            Token[3] = Channel;

            // Set ramp rate
            Token[4] = Ramp;

            // Set new position
            NewPositionInBytes = BitConverter.GetBytes(ServoStart+NewPosition);
            Token[5] = NewPositionInBytes[0];
            Token[6] = NewPositionInBytes[1];

            // Send command to Parallax Servo Controller
            PSCWrite(Token);

            // The command is also returned as a result, delete this data in the buffer
            PSCEmptyReadBuffer();
        }

        public UInt16 GetServoPosition(byte Channel)
            // Read servo position
        {
            byte[] Token = new byte[8];
            byte[] PSCBuffer = new byte[11];

            // The buffer may not be empty, so this is done explicitly
            PSCEmptyReadBuffer();            
            
            //                         01234567  
            Token = Ascii.GetBytes("!SCRSPc\x0D");
            // Syntax: !SCRSP x $0D (no spaces) 
            //         x = servo channel, binary number 1 - 31 (16 - 31 on second board)
            // Reply:  x y z (no spaces / 3 bytes)
            //         z:y = 2 bytes (uint16) for position

            Token[6] = Channel;

            PSCWrite(Token);

            // wait 0.4 seconds to give the PSC the chance to respond
            Pause(0.4);

            // returns the command (8 positions) and x y z (3 positions)
            PSCRead(ref PSCBuffer, 0, 11);

            // High and Low byte have to be in different order for BitConverter.ToUInt16
            PSCBuffer[0] = PSCBuffer[10];
            PSCBuffer[1] = PSCBuffer[9];

            // return only the position
            return (UInt16) (BitConverter.ToUInt16(PSCBuffer, 0) - ServoStart);
        }

    }
}
